/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace BinaryData
{

//================== right_arrow.svg ==================
static const unsigned char temp_binary_data_35[] =
"<svg width=\"128\" height=\"256\" viewBox=\"0 0 128 256\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\">\n"
"<rect x=\"0\" y=\"0\" width=\"100%\" height=\"100%\" fill=\"none\" />\n"
"<path d=\"M0 256L128 128L0 0V256Z\" fill=\"#000000\"/>\n"
"</svg>\n";

const char* right_arrow_svg = (const char*) temp_binary_data_35;
}
