/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace BinaryData
{

//================== highpass.svg ==================
static const unsigned char temp_binary_data_17[] =
"<svg width=\"256\" height=\"256\" viewBox=\"0 0 256 256\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\">\n"
"<rect x=\"0\" y=\"0\" width=\"100%\" height=\"100%\" fill=\"none\" />\n"
"<path d=\"M16 184.5L126.5 74H245\" stroke=\"#000000\" stroke-width=\"30\"/>\n"
"</svg>\n";

const char* highpass_svg = (const char*) temp_binary_data_17;
}
