// Copyright (C) 2026 - zsliu98
// This file is part of ZLEqualizer
//
// ZLEqualizer is free software: you can redistribute it and/or modify it under the terms of the GNU Affero General Public License Version 3 as published by the Free Software Foundation.
//
// ZLEqualizer is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License along with ZLEqualizer. If not, see <https://www.gnu.org/licenses/>.

// This file is also dual licensed under the Apache License, Version 2.0. You may obtain a copy of the License at <http://www.apache.org/licenses/LICENSE-2.0>

// The v1 translation is assisted by the GitHub user https://github.com/ikumi90s.
// Please refer to the v1 commit history for details.

#pragma once

#include <array>

namespace zlpanel::multilingual::ja {
    static constexpr std::array kTexts = {
        "離す：バンドをバイパスします。",
        "押す：バンドをソロにします。",
        "フィルタータイプを選択します。",
        "フィルタースロープを選択します。スロープが高いほど、フィルターのレスポンスカーブの変化が急になります。",
        "ステレオモードを選択します。",
        "周波数を調整します。",
        "ベースゲインとターゲットゲインを調整します。",
        "Qファクターを調整します。Q値が大きいほど、帯域幅が狭くなります。",
        "押す：ダイナミック動作をオンにします。",
        "クリック：バンドをオフにします。",
        "離す：ダイナミック動作をバイパスします。",
        "押す：ダイナミック学習動作をオンにします。\n離す：ダイナミック学習動作をオフにし、閾値とニーを設定します。\n詳細はマニュアルをご覧ください。",
        "押す：ダイナミック相対動作をオンにします。\n詳細はマニュアルをご覧ください。",
        "押す：サイドチェーンのステレオモードを変更します。",
        "ダイナミック動作の閾値を調整します。\n詳細はマニュアルをご覧ください。",
        "ダイナミック動作のニー(Knee)幅を調整します。\n詳細はマニュアルをご覧ください。",
        "ダイナミック動作のアタックタイムを調整します。\n詳細はマニュアルをご覧ください。",
        "ダイナミック動作のリリースタイムを調整します。\n詳細はマニュアルをご覧ください。",
        "押す：バンドをサイドチェーンバンドとリンクします。",
        "サイドチェーンのフィルタータイプを選択します。",
        "サイドチェーンのフィルタースロープを選択します。",
        "サイドチェーンのフィルター周波数を調整します。",
        "サイドチェーンのフィルターQファクターを調整します。",

        "離す：プラグインをバイパスします。",
        "押す：外部サイドチェーンを使用します。\n離す：内部サイドチェーンを使用します。",

        "追加の出力ゲインを調整します。",
        "すべてのフィルターのベース＆ターゲットゲインのスケールを調整します。",
        "押す：静的ゲイン補正をオンにします。SGCは不正確ですが、ダイナミクスには影響しません。",
        "押す：入力信号と出力信号の統合ラウドネスの測定を開始します。\n離す：AGCをオフにし、出力ゲインを2つのラウドネス値の差に更新します。",
        "押す：自動ゲイン補正をオンにします。AGCは長期的にはより正確ですが、ダイナミクスに影響を与えます。",
        "押す：出力信号の位相を反転します。",
        "サイドチェーン信号のルックアヘッドを調整します。",

        "押す：入力信号のスペクトラムアナライザーをオンにします。",
        "押す：出力信号のスペクトラムアナライザーをオンにします。",
        "押す：サイドチェーン信号のスペクトラムアナライザーをオンにします。",
        "スペクトラムアナライザーの減衰速度を選択します。",
        "スペクトラムアナライザーのティルトスロープを選択します。",
        "押す：凍結機能をオンにします。アナライザー上でマウスをホバーするとスペクトラムが凍結されます。",
        "押す：衝突検出をオンにします。",
        "衝突検出の強度を調整します。",

        "フィルター構造を選択します。",
        "標準的でクラシックなデジタル構造です。積極的なオートメーションには適していません。",
        "シンセフィルターやクロスオーバーで使用される安定した構造です。積極的なオートメーションに適しています。大きな位相シフトを引き起こします。",
        "緩やかなシェルフおよびピークフィルターが並列処理されます。効率的で自然なダイナミック処理が可能です。",
        "アナログプロトタイプの振幅＆位相レスポンスを模倣します。このモードではフィルターパラメータをオートメーションしないでください。",
        "アナログプロトタイプの振幅レスポンスを模倣し、高域の位相をクリーンアップします。このモードではフィルターパラメータをオートメーションしないでください。",
        "アナログプロトタイプの振幅レスポンスを模倣し、中高域の位相をクリーンアップします。このモードではフィルターパラメータをオートメーションしないでください。",

        "ダブルクリック：UI設定を開きます。",

        "押す：イコライゼーションマッチパネルを開きます。",
        "クリック：ターゲットカーブをプリセットファイルとして保存します。",
        "選択：ターゲットカーブを選択します。",
        "サイドチェーン信号からターゲットカーブを学習します。",
        "ターゲットカーブをフラットな線として設定します。",
        "プリセットファイルからターゲットカーブを読み込みます。",
        "ターゲットカーブをバランスカーブとして設定します。",
        "ターゲットカーブをナチュラルカーブとして設定します。",
        "差分カーブのシフトを調整します。",
        "差分カーブの滑らかさを調整します。",
        "差分カーブのスロープを調整します。",
        "クリック：フィッティングプロセスを開始します。",
        "バンド数を調整します。",
        "イコライゼーションマッチモデルがメインチェーン信号とサイドチェーン信号を分析しています。\nソースはメインチェーン信号である必要がありますが、ターゲットは選択できます。\nソースとターゲットの差が、差分（ディファレンス）となります。\n差分が安定したら、フィッティングプロセスの前にさらに調整することができます。",
        "イコライゼーションマッチモデルがフィッティングプロセスを実行中です。最大で数秒かかります。",
        "イコライゼーションマッチモデルがフィッティングプロセスを完了しました。フィッティングに使用するバンド数を変更できます。",
        "押す：差分カーブ描画を有効にする。"
    };
}
