/* ==================================== JUCER_BINARY_RESOURCE ====================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#include <cstring>

namespace BinaryData
{

//================== lowpass.svg ==================
static const unsigned char temp_binary_data_19[] =
"<svg width=\"256\" height=\"256\" viewBox=\"0 0 256 256\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\">\n"
"<rect x=\"0\" y=\"0\" width=\"100%\" height=\"100%\" fill=\"none\" />\n"
"<path d=\"M245 184.5L134.5 74H16\" stroke=\"#000000\" stroke-width=\"30\"/>\n"
"</svg>\n";

const char* lowpass_svg = (const char*) temp_binary_data_19;
}
