/* =========================================================================================

   This is an auto-generated file: Any edits you make may be overwritten!

*/

#pragma once

namespace BinaryData
{
    extern const char*   InterSubsetMediumNoHinting_ttf;
    const int            InterSubsetMediumNoHinting_ttfSize = 56440;

    extern const char*   InterSubsetMedium_ttf;
    const int            InterSubsetMedium_ttfSize = 76024;

    extern const char*   bypass_svg;
    const int            bypass_svgSize = 654;

    extern const char*   circle_a_svg;
    const int            circle_a_svgSize = 430;

    extern const char*   circle_r_svg;
    const int            circle_r_svgSize = 568;

    extern const char*   close_svg;
    const int            close_svgSize = 382;

    extern const char*   collision_svg;
    const int            collision_svgSize = 2507;

    extern const char*   dline_a_svg;
    const int            dline_a_svgSize = 408;

    extern const char*   dline_l_svg;
    const int            dline_l_svgSize = 299;

    extern const char*   dline_s_svg;
    const int            dline_s_svgSize = 825;

    extern const char*   draw_svg;
    const int            draw_svgSize = 2413;

    extern const char*   dynamic_svg;
    const int            dynamic_svgSize = 2035;

    extern const char*   externalside_svg;
    const int            externalside_svgSize = 446;

    extern const char*   freeze_svg;
    const int            freeze_svgSize = 385;

    extern const char*   allpass_svg;
    const int            allpass_svgSize = 381;

    extern const char*   bandpass_svg;
    const int            bandpass_svgSize = 413;

    extern const char*   flatshelf_svg;
    const int            flatshelf_svgSize = 292;

    extern const char*   highpass_svg;
    const int            highpass_svgSize = 237;

    extern const char*   highshelf_svg;
    const int            highshelf_svgSize = 331;

    extern const char*   lowpass_svg;
    const int            lowpass_svgSize = 237;

    extern const char*   lowshelf_svg;
    const int            lowshelf_svgSize = 334;

    extern const char*   notch_svg;
    const int            notch_svgSize = 366;

    extern const char*   peak_svg;
    const int            peak_svgSize = 431;

    extern const char*   tiltshelf_svg;
    const int            tiltshelf_svgSize = 440;

    extern const char*   left_arrow_svg;
    const int            left_arrow_svgSize = 220;

    extern const char*   link_svg;
    const int            link_svgSize = 1022;

    extern const char*   left_svg;
    const int            left_svgSize = 286;

    extern const char*   mid_svg;
    const int            mid_svgSize = 406;

    extern const char*   right_svg;
    const int            right_svgSize = 355;

    extern const char*   side_svg;
    const int            side_svgSize = 865;

    extern const char*   stereo_svg;
    const int            stereo_svgSize = 306;

    extern const char*   main_py;
    const int            main_pySize = 2959;

    extern const char*   match_svg;
    const int            match_svgSize = 685;

    extern const char*   phase_svg;
    const int            phase_svgSize = 771;

    extern const char*   reset_settings_svg;
    const int            reset_settings_svgSize = 1139;

    extern const char*   right_arrow_svg;
    const int            right_arrow_svgSize = 218;

    extern const char*   save_svg;
    const int            save_svgSize = 1871;

    extern const char*   shuffle_svg;
    const int            shuffle_svgSize = 392;

    extern const char*   solo_svg;
    const int            solo_svgSize = 777;

    extern const char*   start_svg;
    const int            start_svgSize = 239;

    extern const char*   up_arrow_svg;
    const int            up_arrow_svgSize = 298;

    extern const char*   logo_svg;
    const int            logo_svgSize = 3259;

    extern const char*   zlaudio_svg;
    const int            zlaudio_svgSize = 867;

    // Number of elements in the namedResourceList and originalFileNames arrays.
    const int namedResourceListSize = 43;

    // Points to the start of a list of resource names.
    extern const char* namedResourceList[];

    // Points to the start of a list of resource filenames.
    extern const char* originalFilenames[];

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding data and its size (or a null pointer if the name isn't found).
    const char* getNamedResource (const char* resourceNameUTF8, int& dataSizeInBytes);

    // If you provide the name of one of the binary resource variables above, this function will
    // return the corresponding original, non-mangled filename (or a null pointer if the name isn't found).
    const char* getNamedResourceOriginalFilename (const char* resourceNameUTF8);
}
